/* Have contents take up entire browser window width. */
.wy-nav-content {
    max-width: none !important
}

.wy-side-nav-search .version {
    color: #000000 !important;
}

.wy-side-nav-search>div.switch-menus>div.version-switch select {
    color: #000000;
}

/* override table width restrictions */
.wy-table-responsive table td, .wy-table-responsive table th {
    white-space: normal !important;
}

.wy-table-responsive {
    overflow: visible !important;
}

.custom-code-block {
    border: 1px solid #CCCCCC;
    background-color: #F9F9F9;
    padding: 10px;
    font-family: monospace;
    overflow: auto;
}

/* https://github.com/readthedocs/sphinx_rtd_theme/issues/1301#issuecomment-1876120817 */
.py.property {
    display: block !important;
}
